@@ -262,8 +262,8 @@ dsdata_alloc(u_int32_t len)
 	u_int32_t size;
 
 	size = sizeof(dsdata) + len;
-	x = (dsdata *)malloc(size);
-	memset(x, 0, size);
+	x = (dsdata *)calloc(1, size);
+	if (x == NULL) return NULL;
 
 	x->length = len;
 	x->data = (char *)x + sizeof(dsdata);
