@@ -1071,6 +1071,12 @@ precsize_ntoa(u_int8_t prec)
 				return item;
 			}
 			if (streq(key, "namev6"))
+			{
+				item = [self hostWithKey:val dnstype:ns_t_aaaa];
+				if (item != nil) [item mergeValue:val forKey:"name"];
+				return item;
+			}
+			if (streq(key, "namev46"))
 			{
 				item = [self hostWithKey:val dnstype:ns_t_a];
 				[item mergeValue:val forKey:"name"];
