@@ -34,6 +34,7 @@
  
 #import "Root.h"
 #import "LUGlobal.h"
+#import <sys/time.h>
 
 typedef struct {
 	char *key;
@@ -49,6 +50,9 @@ typedef struct {
 	BOOL negative;
 }
 
+- (LUDictionary *)initTimeStamped;
+- (LUDictionary *)initWithTime:(struct timeval)t;
+
 - (BOOL)isEqual:(LUDictionary *)dict;
 
 - (unsigned int)indexForKey:(char *)key;
@@ -114,6 +118,9 @@ typedef struct {
 - (int)intForKey:(char *)key;
 - (unsigned long)unsignedLongForKey:(char *)key;
 
+- (void)setInt:(int)i forKey:(char *)key;
+- (void)setUnsignedLong:(unsigned long)i forKey:(char *)key;
+
 - (char *)valueForKey:(char *)key;
 - (char *)valueAtIndex:(unsigned int)where;
 
